/* Copyright (c) 2002-2012 Croteam Ltd. 
This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as published by
the Free Software Foundation


This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

// Animation names
#define DEVIL_ANIM_DEFAULT 0
#define DEVIL_ANIM_POSEDOWN 1
#define DEVIL_ANIM_GETUP 2
#define DEVIL_ANIM_ANGER01 3
#define DEVIL_ANIM_ANGER02 4
#define DEVIL_ANIM_CELEBRATE01 5
#define DEVIL_ANIM_PUNCH 6
#define DEVIL_ANIM_SMASH 7
#define DEVIL_ANIM_GRABWEAPONS01 8
#define DEVIL_ANIM_GRABWEAPONS02 9
#define DEVIL_ANIM_FROMIDLETOATTACK01 10
#define DEVIL_ANIM_ATTACK01 11
#define DEVIL_ANIM_FROMATTACK01TOIDLE 12
#define DEVIL_ANIM_FROMIDLETOATTACK02 13
#define DEVIL_ANIM_ATTACK02 14
#define DEVIL_ANIM_FROMATTACK02TOIDLE 15
#define DEVIL_ANIM_ATTACKCLOSE 16
#define DEVIL_ANIM_ATTACKBREATHSTART 17
#define DEVIL_ANIM_ATTACKBREATH 18
#define DEVIL_ANIM_ATTACKBREATHEND 19
#define DEVIL_ANIM_FROMIDLETOWALK 20
#define DEVIL_ANIM_WALK 21
#define DEVIL_ANIM_FROMWALKTOIDLE 22
#define DEVIL_ANIM_IDLE 23
#define DEVIL_ANIM_CLIMB 24
#define DEVIL_ANIM_WOUNDSTART 25
#define DEVIL_ANIM_WOUNDLOOP 26
#define DEVIL_ANIM_WOUNDEND 27
#define DEVIL_ANIM_HEAL 28
#define DEVIL_ANIM_DEATH 29

// Color names

// Patch names

// Names of collision boxes
#define DEVIL_COLLISION_BOX_PART_NAME 0

// Attaching position names
#define DEVIL_ATTACHMENT_LASER 0
#define DEVIL_ATTACHMENT_PROJECTILEGUN 1
#define DEVIL_ATTACHMENT_ELETRICITYGUN 2
#define DEVIL_ATTACHMENT_ROCKETLAUNCHER 3

// Sound names
