/* Copyright (c) 2002-2012 Croteam Ltd. 
This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as published by
the Free Software Foundation


This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

// Animation names
#define HEADMAN_ANIM_DEFAULT_ANIMATION 0
#define HEADMAN_ANIM_FIRECRACKER_ATACK 1
#define HEADMAN_ANIM_ROCKETMAN_ATTACK 2
#define HEADMAN_ANIM_CHAINSAW_ATTACK 3
#define HEADMAN_ANIM_BOMBERMAN_ATTACK 4
#define HEADMAN_ANIM_KAMIKAZE_ATTACK 5
#define HEADMAN_ANIM_CRAWL01 6
#define HEADMAN_ANIM_CRAWLDEATH01 7
#define HEADMAN_ANIM_CRAWLSEARCH02 8
#define HEADMAN_ANIM_DEATH01 9
#define HEADMAN_ANIM_DEATH02 10
#define HEADMAN_ANIM_DEATH03 11
#define HEADMAN_ANIM_DEATH04 12
#define HEADMAN_ANIM_DEATH05 13
#define HEADMAN_ANIM_FALL03 14
#define HEADMAN_ANIM_RUNFIRE01 15
#define HEADMAN_ANIM_RUNFIRE02 16
#define HEADMAN_ANIM_RUNFIRE03 17
#define HEADMAN_ANIM_IDLE01 18
#define HEADMAN_ANIM_IDLE02 19
#define HEADMAN_ANIM_WALK01 20
#define HEADMAN_ANIM_WOUND01 21
#define HEADMAN_ANIM_WOUND02 22
#define HEADMAN_ANIM_WOUND03 23
#define HEADMAN_ANIM_WOUND04 24
#define HEADMAN_ANIM_WOUND05 25
#define HEADMAN_ANIM_MACRODEATH01 26
#define HEADMAN_ANIM_MACRODEATH02 27
#define HEADMAN_ANIM_MACRODEATH03 28
#define HEADMAN_ANIM_MACRODEATH04 29
#define HEADMAN_ANIM_MACRODEATH05 30

// Color names

// Patch names

// Names of collision boxes
#define HEADMAN_COLLISION_BOX_DEFAULT 0
#define HEADMAN_COLLISION_BOX_DEATH 1

// Attaching position names
#define HEADMAN_ATTACHMENT_LEFT_HAND 0
#define HEADMAN_ATTACHMENT_SAW 1
#define HEADMAN_ATTACHMENT_RIGHT_HAND 2
#define HEADMAN_ATTACHMENT_BOMB_1 3
#define HEADMAN_ATTACHMENT_BOMB_2 4
#define HEADMAN_ATTACHMENT_NORMAL_HEAD 5
#define HEADMAN_ATTACHMENT_STICK 6

// Sound names
