/* Copyright (c) 2002-2012 Croteam Ltd. 
This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as published by
the Free Software Foundation


This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

// Animation names
#define PLAYER_ANIM_DEFAULT_ANIMATION 0
#define PLAYER_ANIM_STAND 1
#define PLAYER_ANIM_BACKPEDAL 2
#define PLAYER_ANIM_BACKPEDALRUN 3
#define PLAYER_ANIM_JUMPSTART 4
#define PLAYER_ANIM_JUMPEND 5
#define PLAYER_ANIM_TURNLEFT 6
#define PLAYER_ANIM_TURNRIGHT 7
#define PLAYER_ANIM_RUN 8
#define PLAYER_ANIM_NORMALWALK 9
#define PLAYER_ANIM_CROUCH 10
#define PLAYER_ANIM_RISE 11
#define PLAYER_ANIM_CROUCH_IDLE 12
#define PLAYER_ANIM_CROUCH_TURNLEFT 13
#define PLAYER_ANIM_CROUCH_TURNRIGHT 14
#define PLAYER_ANIM_CROUCH_WALK 15
#define PLAYER_ANIM_CROUCH_WALKBACK 16
#define PLAYER_ANIM_SWIM 17
#define PLAYER_ANIM_SWIMIDLE 18
#define PLAYER_ANIM_DEATH_UNDERWATER 19
#define PLAYER_ANIM_DEATH_UNDERWATERIDLE 20
#define PLAYER_ANIM_DEATH_BACK 21
#define PLAYER_ANIM_DEATH_EASYFALLBACK 22
#define PLAYER_ANIM_DEATH_FORWARD 23
#define PLAYER_ANIM_DEATH_EASYFALLFORWARD 24
#define PLAYER_ANIM_DEATH_SPIKES 25
#define PLAYER_ANIM_STATUE_PUT 26

// Color names

// Patch names

// Names of collision boxes
#define PLAYER_COLLISION_BOX_STAND 0
#define PLAYER_COLLISION_BOX_CROUCH 1
#define PLAYER_COLLISION_BOX_SWIM 2

// Attaching position names
#define PLAYER_ATTACHMENT_BODY 0

// Sound names
