// Animation names
#define KNAPSACK01_ANIM_DEFAULT_ANIMATION 0

// Color names

// Patch names

// Names of collision boxes
#define KNAPSACK01_COLLISION_BOX_PART_NAME 0

// Attaching position names

// Sound names
