/* Copyright (c) 2002-2012 Croteam Ltd. 
This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as published by
the Free Software Foundation


This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

#define _SE_DEMO            0   // set for demo versions
#define _SE_BUILD_MAJOR 10000   // use new number for each released version
#define _SE_BUILD_MINOR    10   // minor versions that are data-compatibile, but are not netgame-compatibile
#define _SE_BUILD_EXTRA    ""   // extra version with minor code changes
#define _SE_VER_STRING  "1.10"  // usually shown in server browser, etc
