/* Copyright (c) 2002-2012 Croteam Ltd. 
This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as published by
the Free Software Foundation


This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

// Animation names
#define ELEMENTALLAVA_ANIM_IDLE 0
#define ELEMENTALLAVA_ANIM_STANDFIGHT 1
#define ELEMENTALLAVA_ANIM_STANDPLANE 2
#define ELEMENTALLAVA_ANIM_ATTACKLEFTHAND 3
#define ELEMENTALLAVA_ANIM_ATTACKTWOHANDS 4
#define ELEMENTALLAVA_ANIM_WOUND01 5
#define ELEMENTALLAVA_ANIM_WOUND02 6
#define ELEMENTALLAVA_ANIM_WOUND03 7
#define ELEMENTALLAVA_ANIM_WALKBIG 8
#define ELEMENTALLAVA_ANIM_RUNSMALL 9
#define ELEMENTALLAVA_ANIM_RUNMEDIUM 10
#define ELEMENTALLAVA_ANIM_ANGER 11
#define ELEMENTALLAVA_ANIM_DEATH03 12
#define ELEMENTALLAVA_ANIM_MELTFLY 13
#define ELEMENTALLAVA_ANIM_MELTUP 14

// Color names

// Patch names

// Names of collision boxes
#define ELEMENTALLAVA_COLLISION_BOX_NORMAL  0
#define ELEMENTALLAVA_COLLISION_BOX_BOX 1

// Attaching position names
#define ELEMENTALLAVA_ATTACHMENT_BODY_FLARE 0
#define ELEMENTALLAVA_ATTACHMENT_RIGHT_HAND_FLARE 1
#define ELEMENTALLAVA_ATTACHMENT_LEFT_HAND_FLARE 2

// Sound names
