/* Copyright (c) 2002-2012 Croteam Ltd. 
This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as published by
the Free Software Foundation


This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

// Animation names
#define LAVAMAN_ANIM_DEFAULTANIMATION 0
#define LAVAMAN_ANIM_ATTACK01 1
#define LAVAMAN_ANIM_ATTACK02 2
#define LAVAMAN_ANIM_ATTACK03 3
#define LAVAMAN_ANIM_ATTACK04 4
#define LAVAMAN_ANIM_ATTACK05 5
#define LAVAMAN_ANIM_ATTACK06 6
#define LAVAMAN_ANIM_DEATH01 7
#define LAVAMAN_ANIM_DEATH02 8
#define LAVAMAN_ANIM_DEATH03 9
#define LAVAMAN_ANIM_STANDBOX 10
#define LAVAMAN_ANIM_MORPHBOXUP 11
#define LAVAMAN_ANIM_MORPHBOXDOWN 12
#define LAVAMAN_ANIM_STANDPLANE 13
#define LAVAMAN_ANIM_MORPHPLANEUP 14
#define LAVAMAN_ANIM_MORPHPLANEDOWN 15
#define LAVAMAN_ANIM_RUN 16
#define LAVAMAN_ANIM_STAND 17
#define LAVAMAN_ANIM_WALK 18
#define LAVAMAN_ANIM_WOUND01 19
#define LAVAMAN_ANIM_WOUND02 20
#define LAVAMAN_ANIM_WOUND03 21

// Color names

// Patch names

// Names of collision boxes
#define LAVAMAN_COLLISION_BOX_NORMAL 0
#define LAVAMAN_COLLISION_BOX_BOX 1

// Attaching position names
#define LAVAMAN_ATTACHMENT_BODY_FLARE 0
#define LAVAMAN_ATTACHMENT_RIGHT_HAND_FLARE 1
#define LAVAMAN_ATTACHMENT_LEFT_HAND_FLARE 2

// Sound names
