/* Copyright (c) 2002-2012 Croteam Ltd. 
This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as published by
the Free Software Foundation


This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

// Animation names
#define DEVIL2_ANIM_DEFAULT 0
#define DEVIL2_ANIM_ANGER01 1
#define DEVIL2_ANIM_ANGER02 2
#define DEVIL2_ANIM_CELEBRATE01 3
#define DEVIL2_ANIM_ATTACKCLOSE 4
#define DEVIL2_ANIM_IDLE 5
#define DEVIL2_ANIM_ALL 6

// Color names

// Patch names

// Names of collision boxes
#define DEVIL2_COLLISION_BOX_default 0
#define DEVIL2_COLLISION_BOX_ 1
#define DEVIL2_COLLISION_BOX_ 2
#define DEVIL2_COLLISION_BOX_ 3
#define DEVIL2_COLLISION_BOX_ 4
#define DEVIL2_COLLISION_BOX_ 5
#define DEVIL2_COLLISION_BOX_ 6
#define DEVIL2_COLLISION_BOX_ 7
#define DEVIL2_COLLISION_BOX_ 8
#define DEVIL2_COLLISION_BOX_ 9
#define DEVIL2_COLLISION_BOX_ 10
#define DEVIL2_COLLISION_BOX_ 11
#define DEVIL2_COLLISION_BOX_ 12
#define DEVIL2_COLLISION_BOX_ 13
#define DEVIL2_COLLISION_BOX_ 14
#define DEVIL2_COLLISION_BOX_ 15
#define DEVIL2_COLLISION_BOX_ 16
#define DEVIL2_COLLISION_BOX_ 17
#define DEVIL2_COLLISION_BOX_ 18
#define DEVIL2_COLLISION_BOX_ 19
#define DEVIL2_COLLISION_BOX_ 20
#define DEVIL2_COLLISION_BOX_ 21
#define DEVIL2_COLLISION_BOX_ 22

// Attaching position names

// Sound names
