/* Copyright (c) 2002-2012 Croteam Ltd. 
This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as published by
the Free Software Foundation


This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

// Animation names
#define CRATEBUS_ANIM_DEFAULT 0
#define CRATEBUS_ANIM_WRECKED 1

// Color names

// Patch names

// Names of collision boxes
#define CRATEBUS_COLLISION_BOX_PART_NAME 0

// Attaching position names
#define CRATEBUS_ATTACHMENT_1 0
#define CRATEBUS_ATTACHMENT_2 1
#define CRATEBUS_ATTACHMENT_3 2
#define CRATEBUS_ATTACHMENT_4 3
#define CRATEBUS_ATTACHMENT_5 4
#define CRATEBUS_ATTACHMENT_6 5
#define CRATEBUS_ATTACHMENT_7 6
#define CRATEBUS_ATTACHMENT_8 7
#define CRATEBUS_ATTACHMENT_9 8
#define CRATEBUS_ATTACHMENT_10 9
#define CRATEBUS_ATTACHMENT_11 10
#define CRATEBUS_ATTACHMENT_12 11
#define CRATEBUS_ATTACHMENT_13 12
#define CRATEBUS_ATTACHMENT_14 13
#define CRATEBUS_ATTACHMENT_15 14
#define CRATEBUS_ATTACHMENT_16 15
#define CRATEBUS_ATTACHMENT_17 16
#define CRATEBUS_ATTACHMENT_18 17
#define CRATEBUS_ATTACHMENT_19 18
#define CRATEBUS_ATTACHMENT_20 19
#define CRATEBUS_ATTACHMENT_21 20
#define CRATEBUS_ATTACHMENT_22 21
#define CRATEBUS_ATTACHMENT_23 22
#define CRATEBUS_ATTACHMENT_24 23

// Sound names
