/* Copyright (c) 2002-2012 Croteam Ltd. 
This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as published by
the Free Software Foundation


This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

// Animation names
#define DRAGONMAN_ANIM_DEFAULT_ANIMATION 0
#define DRAGONMAN_ANIM_AIRATTACKCLOSELOOP 1
#define DRAGONMAN_ANIM_AIRATTACK02 2
#define DRAGONMAN_ANIM_AIRDEATH 3
#define DRAGONMAN_ANIM_AIRFLYLOOP 4
#define DRAGONMAN_ANIM_AIRSTANDLOOP 5
#define DRAGONMAN_ANIM_AIRWOUNDSLIGHT 6
#define DRAGONMAN_ANIM_AIRWOUND02CRITICAL 7
#define DRAGONMAN_ANIM_GROUNDATTACKCLOSELOOP 8
#define DRAGONMAN_ANIM_GROUNDATTACKDISTANT 9
#define DRAGONMAN_ANIM_GROUNDDEATHBACK 10
#define DRAGONMAN_ANIM_GROUNDDEATHFRONT 11
#define DRAGONMAN_ANIM_GROUNDRUN 12
#define DRAGONMAN_ANIM_GROUNDSTANDLOOP 13
#define DRAGONMAN_ANIM_GROUNDTOAIR 14
#define DRAGONMAN_ANIM_AIRTOGROUND 15
#define DRAGONMAN_ANIM_GROUNDWALK 16
#define DRAGONMAN_ANIM_GROUNDWOUNDCRITICALBACK 17
#define DRAGONMAN_ANIM_GROUNDWOUNDCRITICALFRONT 18
#define DRAGONMAN_ANIM_GROUNDWOUNDCRITICALBACK2 19

// Color names

// Patch names

// Names of collision boxes
#define DRAGONMAN_COLLISION_BOX_GROUND 0
#define DRAGONMAN_COLLISION_BOX_AIR 1
#define DRAGONMAN_COLLISION_BOX_DEATH 2

// Attaching position names

// Sound names
