/* Copyright (c) 2002-2012 Croteam Ltd. 
This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as published by
the Free Software Foundation


This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

// Animation names
#define CYBORG_ANIM_DEFAULT_ANIMATION 0
#define CYBORG_ANIM_ATTACKCLOSE01 1
#define CYBORG_ANIM_ATTACKCLOSE02 2
#define CYBORG_ANIM_DEATH01 3
#define CYBORG_ANIM_DEATH02 4
#define CYBORG_ANIM_DEATHREST 5
#define CYBORG_ANIM_DEATHTWIST 6
#define CYBORG_ANIM_TOFIRE 7
#define CYBORG_ANIM_FIRE01 8
#define CYBORG_ANIM_FIRE02 9
#define CYBORG_ANIM_FIRE01SLIDE 10
#define CYBORG_ANIM_FIRE02SLIDE 11
#define CYBORG_ANIM_FROMFIRE 12
#define CYBORG_ANIM_WAIT01 13
#define CYBORG_ANIM_WAIT02 14
#define CYBORG_ANIM_WAIT03 15
#define CYBORG_ANIM_WALK01 16
#define CYBORG_ANIM_WALK02 17
#define CYBORG_ANIM_WOUND01 18
#define CYBORG_ANIM_WOUND02 19
#define CYBORG_ANIM_WOUND03 20
#define CYBORG_ANIM_WOUND04 21
#define CYBORG_ANIM_BIKEREST 22
#define CYBORG_ANIM_FALL01 23
#define CYBORG_ANIM_FALL02 24
#define CYBORG_ANIM_GETUP 25

// Color names

// Patch names

// Names of collision boxes
#define CYBORG_COLLISION_BOX_GROUND 0
#define CYBORG_COLLISION_BOX_DEATH 1
#define CYBORG_COLLISION_BOX_BIKE 2

// Attaching position names
#define CYBORG_ATTACHMENT_ASS 0
#define CYBORG_ATTACHMENT_TORSO 1
#define CYBORG_ATTACHMENT_HEAD 2
#define CYBORG_ATTACHMENT_RIGHTUPPERARM 3
#define CYBORG_ATTACHMENT_RIGHTLOWERARM 4
#define CYBORG_ATTACHMENT_LEFTUPPERARM 5
#define CYBORG_ATTACHMENT_LEFTLOWERARM 6
#define CYBORG_ATTACHMENT_RIGHTUPPERLEG 7
#define CYBORG_ATTACHMENT_RIGHTLOWERLEG 8
#define CYBORG_ATTACHMENT_LEFTUPPERLEG 9
#define CYBORG_ATTACHMENT_LEFTLOWERLEG 10
#define CYBORG_ATTACHMENT_FOOTRIGHT 11
#define CYBORG_ATTACHMENT_FOOTLEFT 12
#define CYBORG_ATTACHMENT_BIKE 13

// Sound names
