/* Copyright (c) 2002-2012 Croteam Ltd. 
This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as published by
the Free Software Foundation


This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

// Animation names
#define FISHMAN_ANIM_DEFAULT_ANIMATION 0
#define FISHMAN_ANIM_GROUNDATTACKLOOP 1
#define FISHMAN_ANIM_GROUNDATTACK02 2
#define FISHMAN_ANIM_GROUNDDEATH 3
#define FISHMAN_ANIM_GROUNDDEATHREST 4
#define FISHMAN_ANIM_GROUNDRUN 5
#define FISHMAN_ANIM_GROUNDSTAND 6
#define FISHMAN_ANIM_GROUNDTOWATER 7
#define FISHMAN_ANIM_GROUNDWALK 8
#define FISHMAN_ANIM_GROUNDWOUND01 9
#define FISHMAN_ANIM_GROUNDWOUND03 10
#define FISHMAN_ANIM_GROUNDWOUND04 11
#define FISHMAN_ANIM_GROUNDWOUND05 12
#define FISHMAN_ANIM_WATERATTACK01 13
#define FISHMAN_ANIM_WATERATTACK02 14
#define FISHMAN_ANIM_WATERDEATH 15
#define FISHMAN_ANIM_WATERDEATHREST 16
#define FISHMAN_ANIM_WATERSTAND 17
#define FISHMAN_ANIM_WATERSWIM01 18
#define FISHMAN_ANIM_WATERSWIM02 19
#define FISHMAN_ANIM_WATERTOGROUND 20
#define FISHMAN_ANIM_WATERWOUND01 21
#define FISHMAN_ANIM_WATERWOUND02 22

// Color names

// Patch names

// Names of collision boxes
#define FISHMAN_COLLISION_BOX_GROUND 0
#define FISHMAN_COLLISION_BOX_WATER 1
#define FISHMAN_COLLISION_BOX_DEATH_GROUND 2
#define FISHMAN_COLLISION_BOX_DEATH_WATER 3

// Attaching position names

// Sound names
