/* Copyright (c) 2002-2012 Croteam Ltd. 
This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as published by
the Free Software Foundation


This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

// Animation names
#define CATMAN_ANIM_DEFAULT_ANIMATION 0
#define CATMAN_ANIM_ATTACK01 1
#define CATMAN_ANIM_ATTACK02 2
#define CATMAN_ANIM_DEATH01 3
#define CATMAN_ANIM_DEATH02 4
#define CATMAN_ANIM_FALL01 5
#define CATMAN_ANIM_FALL02 6
#define CATMAN_ANIM_RUN 7
#define CATMAN_ANIM_STAND 8
#define CATMAN_ANIM_WALK 9
#define CATMAN_ANIM_WOUND01 10
#define CATMAN_ANIM_WOUND02 11
#define CATMAN_ANIM_WOUND03 12

// Color names
#define CATMAN_PART_HOLE ((1L) << 0)

// Patch names

// Names of collision boxes
#define CATMAN_COLLISION_BOX_DEFAULT 0
#define CATMAN_COLLISION_BOX_DEATH 1
#define CATMAN_COLLISION_BOX_LEFT_HAND 2

// Attaching position names
#define CATMAN_ATTACHMENT_KNIFE 0
#define CATMAN_ATTACHMENT_SWORD 1

// Sound names
//sound SOUND_CATMAN_ATTACK01         "Sounds\Test\boom.wav",          // CATMAN_ANIM_ATTACK01, NL, 0.1f
//sound SOUND_CATMAN_ATTACK02         "Sounds\tmp\gun1.wav",           // CATMAN_ANIM_ATTACK02, L, 0.0f
//sound SOUND_CATMAN_DEATH01          "Sounds\tmp\gun1.wav",           // CATMAN_ANIM_DEATH01, NL, 0.0f
