// Animation names
#define MAMUT_ANIM_DEFAULT_ANIMATION 0
#define MAMUT_ANIM_ATTACK01 1
#define MAMUT_ANIM_DEATH 2
#define MAMUT_ANIM_RUN 3
#define MAMUT_ANIM_STAND 4
#define MAMUT_ANIM_WALK 5
#define MAMUT_ANIM_WOUND01 6
#define MAMUT_ANIM_WOUND02 7

// Color names

// Patch names

// Names of collision boxes
#define MAMUT_COLLISION_BOX_DEFAULT 0
#define MAMUT_COLLISION_BOX_DEATH  1

// Attaching position names
#define MAMUT_ATTACHMENT_MAN_FRONT 0
#define MAMUT_ATTACHMENT_MAN_MIDDLE 1
#define MAMUT_ATTACHMENT_MAN_REAR 2

// Sound names
