/* Copyright (c) 2002-2012 Croteam Ltd. 
This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as published by
the Free Software Foundation


This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

// Animation names
#define DEVIL2_ANIM_DEFAULT_ANIMATION 0
#define DEVIL2_ANIM_FROMSTANDTOATTACK01POSITION 1
#define DEVIL2_ANIM_ATTACK01LOOPMINIGUN 2
#define DEVIL2_ANIM_FROMATTACK01TOSTANDPOSITION 3
#define DEVIL2_ANIM_ATTACK02LOOPCLAWS 4
#define DEVIL2_ANIM_FROMSTANDTOATTACK03POSITION 5
#define DEVIL2_ANIM_ATTACK03 6
#define DEVIL2_ANIM_FROMATTACK03TOSTANDPOSITION 7
#define DEVIL2_ANIM_DEATH 8
#define DEVIL2_ANIM_DEATHREST 9
#define DEVIL2_ANIM_FROMSTANDTODEFENDPOSITION 10
#define DEVIL2_ANIM_DEFENDLOOP 11
#define DEVIL2_ANIM_FROMDEFENDTOSTANDPOSITION 12
#define DEVIL2_ANIM_RUN 13
#define DEVIL2_ANIM_STANDLOOP 14
#define DEVIL2_ANIM_WALK 15
#define DEVIL2_ANIM_WOUND01SLIGHTFRONT 16
#define DEVIL2_ANIM_WOUND02SLIGHTBACK 17
#define DEVIL2_ANIM_WOUND03CRITICALFRONT 18

// Color names

// Patch names

// Names of collision boxes
#define DEVIL2_COLLISION_BOX_PART_NAME 0
