/* Copyright (c) 2002-2012 Croteam Ltd. 
This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as published by
the Free Software Foundation


This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

// Animation names
#define HEADMAN_ANIM_DEFAULT_ANIMATION 0
#define HEADMAN_ANIM_COMPUTER 1
#define HEADMAN_ANIM_COMPUTERKAMIKAZE 2
#define HEADMAN_ANIM_FIRECRACKER_ATTACK 3
#define HEADMAN_ANIM_ROCKETMAN_ATTACK 4
#define HEADMAN_ANIM_BOMBERMAN_ATTACK 5
#define HEADMAN_ANIM_KAMIKAZE_ATTACK 6
#define HEADMAN_ANIM_DEATH_FALL_BACK 7
#define HEADMAN_ANIM_DEATH_FALL_ON_KNEES 8
#define HEADMAN_ANIM_DEATH_EASY_FALL_FORWARD 9
#define HEADMAN_ANIM_DEATH_EASY_FALL_BACK 10
#define HEADMAN_ANIM_WOUND1 11
#define HEADMAN_ANIM_WOUND2 12
#define HEADMAN_ANIM_RUN 13
#define HEADMAN_ANIM_IDLE_FIGHT 14
#define HEADMAN_ANIM_IDLE_PATROL 15
#define HEADMAN_ANIM_WALK 16

// Color names

// Patch names

// Names of collision boxes
#define HEADMAN_COLLISION_BOX_DEFAULT 0
#define HEADMAN_COLLISION_BOX_DEATH 1

// Attaching position names
#define HEADMAN_ATTACHMENT_HEAD 0
#define HEADMAN_ATTACHMENT_ROCKET_LAUNCHER 1
#define HEADMAN_ATTACHMENT_CHAINSAW 2
#define HEADMAN_ATTACHMENT_BOMB_RIGHT_HAND 3
#define HEADMAN_ATTACHMENT_BOMB_LEFT_HAND 4

// Sound names
