/* Copyright (c) 2002-2012 Croteam Ltd. 
This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as published by
the Free Software Foundation


This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

// Animation names
#define DEVIL_ANIM_DEFAULT_ANIMATION 0
#define DEVIL_ANIM_FROMSTANDTOATTACK01POSITION 1
#define DEVIL_ANIM_ATTACK01LOOPMINIGUN 2
#define DEVIL_ANIM_FROMATTACK01TOSTANDPOSITION 3
#define DEVIL_ANIM_ATTACK02LOOPCLAWS 4
#define DEVIL_ANIM_ATTACK03 5
#define DEVIL_ANIM_DEATH 6
#define DEVIL_ANIM_DEATHREST 7
#define DEVIL_ANIM_FROMSTANDTODEFENDPOSITION 8
#define DEVIL_ANIM_DEFENDLOOP 9
#define DEVIL_ANIM_FROMDEFENDTOSTANDPOSITION 10
#define DEVIL_ANIM_RUN 11
#define DEVIL_ANIM_STANDLOOP 12
#define DEVIL_ANIM_WALK 13
#define DEVIL_ANIM_WOUND01SLIGHTFRONT 14
#define DEVIL_ANIM_WOUND02SLIGHTBACK 15
#define DEVIL_ANIM_WOUND03CRITICALFRONT 16

// Color names

// Patch names

// Names of collision boxes
#define DEVIL_COLLISION_BOX_DEAFULT 0
#define DEVIL_COLLISION_BOX_DEATH 1

// Attaching position names
#define DEVIL_ATTACHMENT_MINIGUN 0
#define DEVIL_ATTACHMENT_STICK 1
#define DEVIL_ATTACHMENT_SHIELD 2

// Sound names
