/* Copyright (c) 2002-2012 Croteam Ltd. 
This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as published by
the Free Software Foundation


This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

// Animation names
#define BODY_ANIM_FORMIPMODELS 0
#define BODY_ANIM_KNIFE_CROUCHWALK 1
#define BODY_ANIM_KNIFE_CROUCHSTAND 2
#define BODY_ANIM_KNIFE_JUMP_START 3
#define BODY_ANIM_KNIFE_JUMP_END 4
#define BODY_ANIM_KNIFE_PUSH 5
#define BODY_ANIM_KNIFE_PULL 6
#define BODY_ANIM_KNIFE_PUSHPULL 7
#define BODY_ANIM_KNIFE_SWIMMING 8
#define BODY_ANIM_KNIFE_TURNING 9
#define BODY_ANIM_KNIFE_STAND 10
#define BODY_ANIM_KNIFE_WALKRUN 11
#define BODY_ANIM_KNIFE_ATTACK 12
#define BODY_ANIM_MINIGUN_CROUCHWALK 13
#define BODY_ANIM_MINIGUN_CROUCHSTAND 14
#define BODY_ANIM_MINIGUN_JUMP_START 15
#define BODY_ANIM_MINIGUN_JUMP_END 16
#define BODY_ANIM_MINIGUN_PUSH 17
#define BODY_ANIM_MINIGUN_PULL 18
#define BODY_ANIM_MINIGUN_PUSHPULL 19
#define BODY_ANIM_MINIGUN_SWIMMING 20
#define BODY_ANIM_MINIGUN_TURNING 21
#define BODY_ANIM_MINIGUN_WALKRUN 22
#define BODY_ANIM_MINIGUN_STAND 23
#define BODY_ANIM_COLT_CROUCHWALK 24
#define BODY_ANIM_COLT_CROUCHSTAND 25
#define BODY_ANIM_COLT_JUMP_START 26
#define BODY_ANIM_COLT_JUMP_END 27
#define BODY_ANIM_COLT_PUSH 28
#define BODY_ANIM_COLT_PULL 29
#define BODY_ANIM_COLT_PUSHPULL 30
#define BODY_ANIM_COLT_SWIMMING 31
#define BODY_ANIM_COLT_TURNING 32
#define BODY_ANIM_COLT_WALKRUN 33
#define BODY_ANIM_COLT_STAND 34
#define BODY_ANIM_SHOTGUN_CROUCHWALK 35
#define BODY_ANIM_SHOTGUN_CROUCHSTAND 36
#define BODY_ANIM_SHOTGUN_JUMP_START 37
#define BODY_ANIM_SHOTGUN_JUMP_END 38
#define BODY_ANIM_SHOTGUN_PUSH 39
#define BODY_ANIM_SHOTGUN_PULL 40
#define BODY_ANIM_SHOTGUN_PUSHPULL 41
#define BODY_ANIM_SHOTGUN_SWIMMING 42
#define BODY_ANIM_SHOTGUN_TURNING 43
#define BODY_ANIM_SHOTGUN_WALKRUN 44
#define BODY_ANIM_SHOTGUN_STAND 45
#define BODY_ANIM_BODY_DEATH_NORMAL 46
#define BODY_ANIM_BODY_DEATH_UNDERWATER 47
#define BODY_ANIM_BODY_DEATH_SPIKES 48
#define BODY_ANIM_STATUE_PULL 49
#define BODY_ANIM_LOOKAROUNDONCE 50
#define BODY_ANIM_NORMALWALK 51
#define BODY_ANIM_PUSHTOWALK 52
#define BODY_ANIM_WAIT 53
#define BODY_ANIM_PUSHTOWALKCOLT 54
#define BODY_ANIM_PUSHTOWALKKNIFE 55
#define BODY_ANIM_WAITTOLOOKAROUND 56
#define BODY_ANIM_LOOKAROUNDTOWAIT 57
#define BODY_ANIM_KNIFE_PUSH_SLOW 58
#define BODY_ANIM_MINIGUN_PUSH_SLOW 59
#define BODY_ANIM_COLT_PUSH_SLOW 60
#define BODY_ANIM_SHOTGUN_PUSH_SLOW 61
#define BODY_ANIM_LOOKAROUND 62

// Color names

// Patch names

// Names of collision boxes
#define BODY_COLLISION_BOX_STAND 0

// Attaching position names
#define BODY_ATTACHMENT_MINIGUN 0
#define BODY_ATTACHMENT_KNIFE 1
#define BODY_ATTACHMENT_ROCKET_LAUNCHER 2
#define BODY_ATTACHMENT_GRENADE_LAUNCHER 3
#define BODY_ATTACHMENT_LASER 4
#define BODY_ATTACHMENT_COLT_RIGHT 5
#define BODY_ATTACHMENT_COLT_LEFT 6
#define BODY_ATTACHMENT_DOUBLE_SHOTGUN 7
#define BODY_ATTACHMENT_SINGLE_SHOTGUN 8
#define BODY_ATTACHMENT_GHOST_BUSTER 9
#define BODY_ATTACHMENT_FLAMER 10
#define BODY_ATTACHMENT_TOMMYGUN 11
#define BODY_ATTACHMENT_PIPE_STICK 12
#define BODY_ATTACHMENT_PIPE_BOMB 13
#define BODY_ATTACHMENT_CANNON 14
#define BODY_ATTACHMENT_HEAD 15
#define BODY_ATTACHMENT_ITEM 16

// Sound names
