/* Copyright (c) 2002-2012 Croteam Ltd. 
This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as published by
the Free Software Foundation


This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

// Animation names
#define BODY_ANIM_DEFAULT 0

// Color names

// Patch names

// Names of collision boxes
#define BODY_COLLISION_BOX_PART_NAME 0

// Attaching position names
#define BODY_ATTACHMENT_FLARE1 0
#define BODY_ATTACHMENT_FLARE2 1
#define BODY_ATTACHMENT_FLARE3 2
#define BODY_ATTACHMENT_FLARE4 3

// Sound names
