/* Copyright (c) 2002-2012 Croteam Ltd. 
This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as published by
the Free Software Foundation


This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

// Animation names
#define EYEMAN_ANIM_DEFAULT_ANIMATION 0
#define EYEMAN_ANIM_MORPHUP 1
#define EYEMAN_ANIM_MORPHDOWN 2
#define EYEMAN_ANIM_ATTACK02 3
#define EYEMAN_ANIM_MORPHATTACKFLY 4
#define EYEMAN_ANIM_MORPHATTACK 5
#define EYEMAN_ANIM_DEATH01 6
#define EYEMAN_ANIM_DEATH02 7
#define EYEMAN_ANIM_MORPHDEATH 8
#define EYEMAN_ANIM_FALL01 9
#define EYEMAN_ANIM_FALL02 10
#define EYEMAN_ANIM_STAND 11
#define EYEMAN_ANIM_WOUND03 12
#define EYEMAN_ANIM_WOUND06 13
#define EYEMAN_ANIM_WOUND07 14
#define EYEMAN_ANIM_MORPHWOUND01 15
#define EYEMAN_ANIM_MORPHWOUND02 16
#define EYEMAN_ANIM_WALK 17
#define EYEMAN_ANIM_RUN 18

// Color names

// Patch names

// Names of collision boxes
#define EYEMAN_COLLISION_BOX_GROUND 0
#define EYEMAN_COLLISION_BOX_AIR 1
#define EYEMAN_COLLISION_BOX_DEATH 2

// Attaching position names

// Sound names
