/* Copyright (c) 2002-2012 Croteam Ltd. 
This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as published by
the Free Software Foundation


This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

// Animation names
#define MAMUTMAN_ANIM_ATTACK01 0
#define MAMUTMAN_ANIM_ATTACK02 1
#define MAMUTMAN_ANIM_DEATH01 2
#define MAMUTMAN_ANIM_DEATH02 3
#define MAMUTMAN_ANIM_FALL01 4
#define MAMUTMAN_ANIM_GETUP01 5
#define MAMUTMAN_ANIM_FALL02 6
#define MAMUTMAN_ANIM_GETUP02 7
#define MAMUTMAN_ANIM_RUN 8
#define MAMUTMAN_ANIM_STAND 9
#define MAMUTMAN_ANIM_WALK 10
#define MAMUTMAN_ANIM_WOUND01 11
#define MAMUTMAN_ANIM_WOUND02 12
#define MAMUTMAN_ANIM_WOUND03 13
#define MAMUTMAN_ANIM_STANDMOUNTEDFIRST 14
#define MAMUTMAN_ANIM_STANDMOUNTEDSECOND 15
#define MAMUTMAN_ANIM_STANDMOUNTEDTHIRD 16
#define MAMUTMAN_ANIM_MOUNTEDATTACKFIRST 17
#define MAMUTMAN_ANIM_MOUNTEDATTACKSECOND 18
#define MAMUTMAN_ANIM_MOUNTEDATTACKTHIRD 19
#define MAMUTMAN_ANIM_FALLFROMMAMUTFIRST 20
#define MAMUTMAN_ANIM_FALLFROMMAMUTSECOND 21
#define MAMUTMAN_ANIM_FALLFROMMAMUTTHIRD 22
#define MAMUTMAN_ANIM_MOUNTEDRUNFIRST 23
#define MAMUTMAN_ANIM_MOUNTEDRUNSECOND 24
#define MAMUTMAN_ANIM_MOUNTEDRUNTHIRD 25
#define MAMUTMAN_ANIM_MOUNTEDWALKFIRST 26
#define MAMUTMAN_ANIM_MOUNTEDWALKSECOND 27
#define MAMUTMAN_ANIM_MOUNTEDWALKTHIRD 28

// Color names

// Patch names

// Names of collision boxes
#define MAMUTMAN_COLLISION_BOX_DEFAULT 0
#define MAMUTMAN_COLLISION_BOX_DEATH 1

// Attaching position names

// Sound names
