/* Copyright (c) 2002-2012 Croteam Ltd. 
This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as published by
the Free Software Foundation


This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

// Animation names
#define BODY_ANIM_DEFAULT_ANIMATION 0
#define BODY_ANIM_COLT_STAND 1
#define BODY_ANIM_COLT_REDRAW 2
#define BODY_ANIM_COLT_DRAW 3
#define BODY_ANIM_COLT_FIRERIGHT 4
#define BODY_ANIM_COLT_FIRELEFT 5
#define BODY_ANIM_KNIFE_ATTACK 6
#define BODY_ANIM_COLT_SWIM_STAND 7
#define BODY_ANIM_COLT_SWIM_REDRAW 8
#define BODY_ANIM_COLT_SWIM_DRAW 9
#define BODY_ANIM_COLT_SWIM_FIRERIGHT 10
#define BODY_ANIM_COLT_SWIM_FIRELEFT 11
#define BODY_ANIM_KNIFE_SWIM_ATTACK 12
#define BODY_ANIM_SHOTGUN_STAND 13
#define BODY_ANIM_SHOTGUN_REDRAW 14
#define BODY_ANIM_SHOTGUN_DRAW 15
#define BODY_ANIM_SHOTGUN_FIRELONG 16
#define BODY_ANIM_SHOTGUN_FIRESHORT 17
#define BODY_ANIM_SHOTGUN_SWIM_STAND 18
#define BODY_ANIM_SHOTGUN_SWIM_REDRAW 19
#define BODY_ANIM_SHOTGUN_SWIM_DRAW 20
#define BODY_ANIM_SHOTGUN_SWIM_FIRELONG 21
#define BODY_ANIM_SHOTGUN_SWIM_FIRESHORT 22
#define BODY_ANIM_MINIGUN_STAND 23
#define BODY_ANIM_MINIGUN_REDRAW 24
#define BODY_ANIM_MINIGUN_DRAW 25
#define BODY_ANIM_MINIGUN_FIRELONG 26
#define BODY_ANIM_MINIGUN_FIRESHORT 27
#define BODY_ANIM_MINIGUN_SWIM_STAND 28
#define BODY_ANIM_MINIGUN_SWIM_REDRAW 29
#define BODY_ANIM_MINIGUN_SWIM_DRAW 30
#define BODY_ANIM_MINIGUN_SWIM_FIRELONG 31
#define BODY_ANIM_MINIGUN_SWIM_FIRESHORT 32
#define BODY_ANIM_DEATH_UNDERWATER 33
#define BODY_ANIM_DEATH_UNDERWATERIDLE 34
#define BODY_ANIM_DEATH_BACK 35
#define BODY_ANIM_DEATH_EASYFALLBACK 36
#define BODY_ANIM_DEATH_FORWARD 37
#define BODY_ANIM_DEATH_EASYFALLFORWARD 38
#define BODY_ANIM_DEATH_SPIKES 39
#define BODY_ANIM_COLT_REDRAWSLOW 40
#define BODY_ANIM_COLT_DEACTIVATETOWALK 41
#define BODY_ANIM_SHOTGUN_REDRAWSLOW 42
#define BODY_ANIM_SHOTGUN_DEACTIVATETOWALK 43
#define BODY_ANIM_MINIGUN_REDRAWSLOW 44
#define BODY_ANIM_MINIGUN_DEACTIVATETOWALK 45
#define BODY_ANIM_STATUE_PULL 46
#define BODY_ANIM_LOOKAROUNDONCE 47
#define BODY_ANIM_WAIT 48
#define BODY_ANIM_WAITTOLOOKAROUND 49
#define BODY_ANIM_LOOKAROUNDTOWAIT 50
#define BODY_ANIM_LOOKAROUND 51
#define BODY_ANIM_NORMALWALK 52
#define BODY_ANIM_NORMALRUN 53

// Color names

// Patch names

// Names of collision boxes
#define BODY_COLLISION_BOX_STAND 0

// Attaching position names
#define BODY_ATTACHMENT_HEAD 0
#define BODY_ATTACHMENT_COLT_RIGHT 1
#define BODY_ATTACHMENT_KNIFE 2
#define BODY_ATTACHMENT_SINGLE_SHOTGUN 3
#define BODY_ATTACHMENT_DOUBLE_SHOTGUN 4
#define BODY_ATTACHMENT_FLAMER 5
#define BODY_ATTACHMENT_TOMMYGUN 6
#define BODY_ATTACHMENT_CANNON 7
#define BODY_ATTACHMENT_GRENADE_LAUNCHER 8
#define BODY_ATTACHMENT_LASER 9
#define BODY_ATTACHMENT_MINIGUN 10
#define BODY_ATTACHMENT_ROCKET_LAUNCHER 11
#define BODY_ATTACHMENT_COLT_LEFT 12
#define BODY_ATTACHMENT_ITEM 13

// Sound names
